#ifndef BASE_SYSTEM_H
#define BASE_SYSTEM_H

#include "dbg.h"
#include "detect.h"
#include "fs.h"
#include "io.h"
#include "mem.h"
#include "secure.h"
#include "str.h"
#include "time.h"
#include "types.h"

#endif
