// This file can be included several times.

#ifndef MAPBUG
// This helps IDEs properly syntax highlight the uses of the macro below.
#define MAPBUG(Id, Name)
#endif

MAPBUG(BUG_GRENADE_DOUBLEEXPLOSION, "grenade-doubleexplosion@ddnet.tw")
